#include<qapplication.h>
#include<qwidget.h>
#include<qwmatrix.h>
#include<qdatetime.h>
#include<qimage.h>
#include<qbitmap.h>
#include<qbrush.h>
#include<qpainter.h>

#include<stdio.h>
#include<stdlib.h>
#include <unistd.h>
#include <ctype.h>


// pixmaps:  object, raster
#include "p_object.xpm"
#include "p_raster.xpm"

int num=1; bool useMask=TRUE; char *cmd=NULL;
const char *pixmapName=NULL, *rasterName=NULL, *rasterColor=NULL;
int delay=5; bool bNoFlicker=TRUE; int distance=-1;
static const QPoint p0(0,0);

inline int max(int x,int y) {return (x>y) ? x : y; }

// if you want to backport this prog to Qt-1.x, then just throw 
// away the trailing ": Qt" in the following line:
class Data : Qt
{
public:
    Data();
public:
    struct Desk:public QWidget{Desk();} _desktop; QRect _deskRect;
    QPixmap _object,_raster,_rotObj[4],_rotInv[4],_r2x2; 
    QPoint _objCent; 
};
Data::Desk::Desk():QWidget(0,0,WType_Desktop|WPaintDesktop|WPaintUnclipped){}

Data::Data()
{
    QBitmap bm;
    // read Pixmaps.
    if (pixmapName) {
        _object.load(pixmapName,"XPM");
    } else {
        _object=QPixmap((const char **)object);
    }
    if (!useMask) { 
        bm.resize(_object.size()); 
        bm.fill(color1);  // do NOT use white or black here !
        _object.setMask(bm);
    }
    if (rasterName) {
        _raster.load(rasterName,"XPM");
    } else {
        _raster=QPixmap((const char **)raster);
    }
    if (!useMask) { 
        bm.resize(_raster.size()); 
        bm.fill(color1);  // do NOT use white or black here !
        _raster.setMask(bm);
    }
    if (_raster.isNull()) {
        fprintf(stderr,"%s: invalid raster pixmap.\n",cmd);
        exit(2);
    }
    if (_object.isNull()) {
        fprintf(stderr,"%s: invalid object pixmap.\n",cmd);
        exit(2);
    }
    if (_object.width() != _object.height()) {
        fprintf(stderr,"%s: object pixmap not square.\n",cmd);
        exit(2);
    }
    _objCent= QPoint( _object.width()/2, _object.height()/2);
    if (distance < 0) { 
        distance= (delay>0) ? _object.width()*2 : 0; distance *= distance;
    }

    QWMatrix m; m.rotate(-90);
    for (int i=0; i<4; i++) {
        _rotObj[i]=_object.xForm(m); m.rotate(90);
        _rotInv[i]=_rotObj[i]; _rotInv[i].detach();
        QBitmap tbm( _rotObj[i].size() ); 
        tbm.fill(color1);   // do NOT use white or black here !

        // at this place there used to be a workaround for 
        //  a bug in some (old) windoze-version of Qt, which I hope
        //  should be fixed by now. if not, tell me, or better 
        //  complain to the trolls. ;-)

        bitBlt(&tbm,p0,_rotObj[i].mask(),tbm.rect(),XorROP);
        _rotInv[i].fill(color0);    // do NOT use white or black here !
        _rotInv[i].setMask(tbm);
    }
    QRect rr=_raster.rect(), cr=_object.rect(); QSize mr;

    //_r2x2=QPixmap(rr.width()*2,rr.height()*2);
    mr.setWidth(max(rr.width(),cr.width())+cr.width());
    mr.setHeight(max(rr.height(),cr.height())+cr.height());

    _r2x2=QPixmap(mr); 
    if (rasterColor) {
        _r2x2.fill(QColor(rasterColor));
    } else { 
        QBrush br(black,_raster); 
        QPainter gc(&_r2x2);
        gc.fillRect(_r2x2.rect(),br);
    }

    _deskRect=_desktop.rect(); 
    _deskRect.setRight( _desktop.width() -_object.width() );
    _deskRect.setBottom( _desktop.height() -_object.height() );
}

class Object : public QObject
{
public:
    Object(Data *d,QObject *parent, const char *name=0 );
    ~Object();
protected:
    Data *_d;  QPoint _pos,_oldpos; 
    int _dir, _olddir, _speed, _step, _timerId;
protected:
    void timerEvent(QTimerEvent *e);
    void drawObject();
    void turnObject();
    void turnTo(const QPoint &);
    void turnFrom(const QPoint &);
    void moveObject();
    void clearObject(bool bOld=TRUE);
    void clearPxmObject(QPixmap *pPM);
    bool checkRaster();
    bool checkMouse(const QPoint&);
    void changeSpeed(bool fast);
    void compressPixmap(QPixmap &pm);
};

Object::Object(Data *d,QObject *parent, const char *name )
    : QObject( parent, name ) , _d(d)
{
    QRect &dR = _d->_deskRect; _timerId=0; _step = 1 ;

    // initial position, speed, ... 
    _dir = rand() % 4;  changeSpeed(TRUE);
    if (_dir & 1) { // up/down
        _pos.setX( dR.left() + rand()% dR.width() );
        _pos.setY( (_dir >= 2) ? dR.top() : dR.bottom() ); 
    } else { // left/right
        _pos.setY( dR.top() + rand()% dR.height() );
        _pos.setX( (_dir >= 2) ? dR.left() : dR.right() ); 
    }
}

Object::~Object() { clearObject(FALSE); }

void Object::clearObject(bool bOld) 
{
    QPoint p=(bOld)?_oldpos:_pos; 
    int dir=(bOld)?_olddir:_dir;
    QSize rs=_d->_raster.size(), cs=_d->_object.size();
    QPoint rpos( p.x() % rs.width() , p.y() % rs.height() );
    QPixmap tmp(cs);
    bitBlt(&tmp,p0,&_d->_r2x2,QRect(rpos,cs),CopyROP);
    tmp.setMask(*_d->_rotObj[dir].mask() );
    bitBlt(&_d->_desktop,p,&tmp,tmp.rect(),CopyROP);
}
void Object::clearPxmObject(QPixmap *pPM) 
{
    QSize rs=_d->_raster.size(), cs=_d->_object.size();
    QPoint rpos( _oldpos.x() % rs.width() , _oldpos.y() % rs.height() );
    QPoint dpos( _oldpos.x() - _pos.x() , _oldpos.y() - _pos.y() );
    QPixmap tmp(cs);
    bitBlt(&tmp,p0,&_d->_r2x2,QRect(rpos,cs),CopyROP);
    tmp.setMask(*_d->_rotObj[_dir].mask() );
    bitBlt( pPM, dpos, &tmp , tmp.rect() , CopyROP );
}
void Object::drawObject()
{
    bitBlt( &_d->_desktop, _pos, 
            &_d->_rotObj[_dir], _d->_rotObj[_dir].rect(), CopyROP );
}
void Object::timerEvent(QTimerEvent *)
{
    _oldpos=_pos; _olddir=_dir;
    if (bNoFlicker) {
        moveObject();
        clearObject();
    } else {
        clearObject();
        moveObject();
    }
    drawObject();
}

bool Object::checkMouse(const QPoint &mp)
{
    QPoint d(mp - _pos);
    return (d.x()*d.x()+d.y()*d.y() >= distance);
}

bool Object::checkRaster()
{
    QSize rs=_d->_raster.size(), cs=_d->_object.size();
    QPoint rpos( _pos.x() % rs.width() , _pos.y() % rs.height() );

    QPixmap ps(cs); 
    // get new object's rectangle from screen
    bitBlt(&ps,p0,&_d->_desktop,QRect(_pos,cs),CopyROP);
    // eventually remove relicts of old object now. 
    //   otherwise it has been done directly on screen before.
    if (bNoFlicker) { clearPxmObject(&ps); }
    // check against the raster
    bitBlt(&ps,p0,&_d->_r2x2,QRect(rpos,cs),XorROP);

    // filter away all that is outside of the new object
    if (useMask) { bitBlt(&ps,p0,&_d->_rotInv[_dir],ps.rect(),CopyROP); }

    // fold image: reduce size of image at the cost of two extra bitblt()'s.
    // (we're only interested whether there are non-0 pixels in it)
    compressPixmap(ps);

    // now search for non-black spots:
    QImage qr=ps.convertToImage();
    if (qr.depth()<8) { qr=qr.convertDepth(8); }
    if (qr.depth()==8) { 
        char *uptr=(char*)qr.bits(), *end=uptr+(qr.numBytes()); 
        while (uptr<end) { if (*uptr) { return FALSE; } uptr++; }
    } else { if (qr.depth()<32) { qr=qr.convertDepth(32); }
        QRgb *uptr=(QRgb*)qr.bits(), *end=uptr+(qr.numBytes()/4); 
        while (uptr<end) { if (*uptr&RGB_MASK) { return FALSE; } uptr++; }
    }
    return TRUE;
}
void Object::compressPixmap(QPixmap &pm) {
    int w=pm.width(), h=pm.height();
    if (w&1 || h&1 || w<8 || h<8) return;
    bitBlt(&pm,0,0,&pm,0,h/2,w,h/2,OrROP);
    bitBlt(&pm,0,0,&pm,w/2,0,w/2,h/2,OrROP);
    pm.resize(w/2,h/2);
}

void Object::turnObject() { _dir += 3 + rand() % 3 ; _dir %= 4; }

void Object::turnTo(const QPoint &tgt) { // care: tgt != _pos 
    QPoint d(tgt-_pos); 
    if (rand()%(abs(d.x())+abs(d.y()))<abs(d.x())) {
        _dir = (d.x()>0) ? 2 : 0 ;
    } else {
        _dir = (d.y()>0) ? 3 : 1 ;
    }
}
void Object::turnFrom(const QPoint &tgt) {
    QPoint d(tgt-_pos); 
    if (tgt==_pos) { _dir = (_dir + 2) % 4; return; }
    if (rand()%(abs(d.x())+abs(d.y()))<abs(d.x())) {
        _dir = (d.x()>0) ? 0 : (d.x()<0) ? 2 : (_dir + 2) % 4;
    } else {
        _dir = (d.y()>0) ? 1 : (d.y()<0) ? 3 : (_dir + 2) % 4;
    }
}


void Object::changeSpeed(bool fast) {
    int bd= (fast) ? delay : 5*delay ;
    _speed = bd + rand() % (bd+1); 
    if (fast) { _step++; if (_step>7) _step=7; } else { _step=1; }
    if (_timerId) { killTimer(_timerId); }
    _timerId=startTimer(_speed);
}

void Object::moveObject()
{
    QPoint oldpos(_pos), mpos; 
    switch (_dir) {
        case 0: _pos.rx() -= _step; break; // left
        case 1: _pos.ry() -= _step; break; // up
        case 2: _pos.rx() += _step; break; // right
        case 3: _pos.ry() += _step; break; // down
    }
    if ( ! _d->_deskRect.contains(_pos)) { 
        _pos=oldpos; turnObject(); changeSpeed(FALSE); return; 
    }
    if ( distance && !checkMouse(mpos=QCursor::pos()-_d->_objCent)) { 
        // chasemode and too near to mouse
        turnFrom(mpos); 
    } else if ( ! checkRaster()) { // not same as raster ...
        if (_step>1) {
            clearObject(FALSE); _step/=2; 
        } else {
            clearObject(FALSE); _pos=oldpos; 
            changeSpeed(FALSE); turnObject(); 
        }
        return;
    } else if ( (rand() % 100) < 2 ) {
        turnObject();
    }
    changeSpeed(TRUE);
}

void usage()
{
    fprintf(stderr, "\
usage: %s [-n #] [-d #] [-c #] [-m] [-p <file>] [-r <file>]\n\
   -n #  number of pacmen (default=%d)\n\
   -d #  minimum delay between moves. (default=%d)\n\
   -p <file>   alternative xpm-file for the pacman (must be square)\n\
   -r <file>   xpm-file for background (what pacman will leave behind)\n\
   -R <color>  solid color for background (what pacman will leave behind)\n\
   -c #  set \"reach\" of chase-mode: pacman will flee away from mouse \n\
            until the supplied distance is reached. (If distance is \n\
            set to 0  then chase-mode is turned off.) \n\
   -m    disable transparent mode (try this, if pacman doesn't move)\n\
   -f    disable flicker-free mode (saves a little bandwidth) \n\
", cmd, num, delay);
}

int main(int argc,char **argv) {
    QApplication q(argc,argv); 
    cmd=q.argv()[0]; 
    
    const QTime &tm=QTime::currentTime();
    srand( tm.msec() + 1000* tm.second() +
        60000* tm.minute() + 3600000*tm.hour() );
    
    for (int i=1; i<qApp->argc(); i++) {
        const char *cptr=qApp->argv()[i];
        if (*cptr!='-') { usage(); exit(1); }
        switch (cptr[1]) {
            case 'n': 
                if (cptr[2]!=0) {cptr+=2;} 
                else {cptr=qApp->argv()[i+1];i++;} 
                if (!cptr) { cptr=""; }
                num=atoi(cptr); if (num<1) {num=1;}
                break;
            case 'd': 
                if (cptr[2]!=0) {cptr+=2;} 
                else {cptr=qApp->argv()[i+1];i++;} 
                if (!cptr) { cptr=""; }
                delay=atoi(cptr); if (delay<0) {delay=0;}
                break;
            case 'c': 
                if (cptr[2]!=0) {cptr+=2;} 
                else { cptr=qApp->argv()[i+1]; i++; }
                if (!cptr) { cptr=""; }
                distance=atoi(cptr); distance *= distance; 
                break;
            case 'p':
                if (cptr[2]!=0) {cptr+=2;} 
                else {cptr=qApp->argv()[i+1];i++;} 
                if (!cptr) { cptr=""; }
                pixmapName=cptr;
                break;
            case 'R':
                if (cptr[2]!=0) {cptr+=2;} 
                else {cptr=qApp->argv()[i+1];i++;} 
                if (!cptr) { cptr=""; }
                rasterColor=cptr;
                break;
            case 'r':
                if (cptr[2]!=0) {cptr+=2;} 
                else {cptr=qApp->argv()[i+1];i++;} 
                if (!cptr) { cptr=""; }
                rasterName=cptr;
                break;
            case 'm':
                useMask=FALSE;
                break;
            case 'f':
                bNoFlicker=FALSE;
                break;
            default: usage(); exit(1);
        } 
    }

    // argv must have been parsed before instanciating Data
    Data d; 

    {
        QObject lifetime;
        for (int i=0; i< num; i++) { 
            new Object(&d,&lifetime); 
        }
        qApp->exec();
    }
    
    return 0;
}

